ScriptName SLPGNUtils extends Quest
;  The home of various utility functions.
SLPGN_MCMConfig Property Settings Auto
SLPGNDeflateOnDeath Property deflation Auto hidden
SexlabFramework property SexLab auto

Actor Property PlayerRef Auto
Actor[] currentFollowers
Faction Property CalmFaction Auto
SexLabPlayerNecroScript Property Main Auto

ImageSpaceModifier Property FadeEffect Auto

Spell RAAS_BreathSpell
Spell WetCold_Spell
Spell Facelight_Spell
Spell FacelightDark_Spell
Spell CondiExp_Effects

Spell EBTBlood_Spell

int[] Property MfgModifierStrength Auto
int[] Property MfgPhonemeStrength Auto
int[] Property MfgExpression Auto


; ---------------------------------- Dealing with the very life of the player character --------------------------------------------------------------------------

Function ResurrectPlayer()
    If(Settings.GetUseOldResurrection())
        PlayerRef.Resurrect()
    Else    
        PlayerRef.SetUnconscious(true)
        Utility.Wait(0.1) 
        PlayerRef.ResetHealthAndLimbs()
        Utility.Wait(0.1) 
        PlayerRef.SetUnconscious(false)
        PlayerRef.PushActorAway(PlayerRef, 0)
        While(PO3_SKSEFunctions.GetActorKnockState(PlayerRef)==5)
            Utility.Wait(0.1) 
        EndWhile
        DoMenus()
    EndIf
EndFunction


Function KillPlayer()
    log("Killing player silently")
    Float i = 2.0   
    
    While (!PlayerRef.IsDead() && (i > 0.0))    
        PlayerRef.KillSilent()      
        Utility.Wait(0.5)       
        i -= 0.5        
    EndWhile

    ;UnCalmActors(Main.GetActorsAround())

;   If(!Settings.GetNaturalResurrection())  
    Main.RandomTrigger()
;   Else
;       NaturalResurrection()
;   EndIf       
EndFunction

; ---------------------------------- These, in theory should keep nearby NPCs calm during the necro acts. It doesn't always work though. -------------------------

Function CalmActors(Actor victim, Actor[] actorsAround)
    Int j = 0
    victim.AddToFaction(CalmFaction)
    PlayerRef.StopCombat()
    PlayerRef.StopCombatAlarm()

    while (j < actorsAround.length)
        if (actorsAround[j])
            actorsAround[j].AddToFaction(CalmFaction)
        endIf
        j += 1
    endWhile
EndFunction


Function UnCalmActors(Actor[] actorsAround)
    int j = 0
    Utility.Wait(2.0)
    PlayerRef.RemoveFromFaction(CalmFaction)
        
    While (j < actorsAround.Length) 
        If (actorsAround[j])        
            actorsAround[j].RemoveFromFaction(CalmFaction)          
        EndIf
        j += 1  
    EndWhile
EndFunction

; ---------------------------------- Dealing with the follower of the player character. --------------------------------------------------------------------------
Function KillFollowers()
    If(Settings.GetKillFollowers())
        currentFollowers = PO3_SKSEFunctions.GetPlayerFollowers() ; Get followers via Powerofthree SKSE
        
        int followerCount = 0
        If(currentFollowers.Find(None) > 0)
            followerCount = currentFollowers.Find(None) 
        ElseIf (currentFollowers.Find(None) == -1)
            followerCount = currentFollowers.Length
        Else
            return
        EndIf
        
        int i = 0
        While (i < followerCount)
            ActorBase followerBase = currentFollowers[i].getActorBase()
            followerBase.setEssential(true)
            followerBase.setEssential(false)
            currentFollowers[i].killEssential()
            if(Settings.GetEnableExpression())
                ApplyDeadExpression(currentFollowers[i])
            endif
            i += 1
        EndWhile
    EndIf
EndFunction

; ========================================= ; Dismissal is currently disabled in favor of killing the followers instantly
;Function DismissFollower()
;    int FollowerFramework = Main.GetFollowerFramework()
;    If(FollowerFramework == 0)
;        Form DialogueFollowerQuest = Game.GetFormFromFile(0x000750BA, "Skyrim.esm")
;        (DialogueFollowerQuest as DialogueFollowerScript).DismissFollower()
;    ElseIf(FollowerFramework == 1)
;        Form NFFFollowerController = Game.GetFormFromFile(0x0000434F, "nwsFollowerFramework.esp")
;        (NFFFollowerController as nwsFollowerControllerScript).DismissAll()
;    EndIf
;    
;EndFunction


Function ResurrectFollower(Actor follower)
    follower.SetUnconscious(true)
    follower.ResetHealthAndLimbs()
    follower.SetUnconscious(false)
    follower.PushActorAway(follower, 0)
EndFunction

; Note that at least NFF has a protected mode for followers, so "killessential" may be needed when it comes time. 
Function KillFollower(Actor follower)
    log("Killing follower silently")
    Float i = 2.0   
    
    While (!follower.IsDead() && (i > 0.0))    
        follower.KillSilent()      
        Utility.Wait(0.5)       
        i -= 0.5        
    EndWhile
    
    ; Reapply death expression after killing
    ApplyDeadExpression(follower)
EndFunction

; ---------------------------------- Expression Handling -----------------------------------------------------------------------------------------

Function ApplyDeadExpression(Actor victim)
    if(Settings.GetEnableExpression())
        Form ConditionalExpression = Game.GetFormFromFile(0x00000D64, "Conditional Expressions.esp")
        if(ConditionalExpression && victim == PlayerRef)
        ;       Debug.Notification("Conditional Expressions detected, removing spell from player")
            CondiExp_Effects = Game.GetFormFromFile(0x00000D61, "Conditional Expressions.esp") as Spell
            PlayerRef.RemoveSpell(CondiExp_Effects)
        endIf
        ;        Debug.Notification("Applying expression to " + victim.GetActorBase().getName())
        Utility.Wait(0.2) 
        MfgPhonemeStrength = Settings.GetMfgPhonemeStrength()
        MfgModifierStrength = Settings.GetMfgModifierStrength()
        MfgExpression = Settings.GetMfgExpression()
        
        int x = 0
        while (x < 16)
            MfgConsoleFunc.SetPhoneme(victim, x, MfgPhonemeStrength[x])
            MfgConsoleFunc.SetModifier(victim, x, MfgModifierStrength[x])
            if(MfgExpression[x] > 0)
                victim.SetExpressionOverride(x, MfgExpression[x])
            endif
            x += 1
        endwhile
    endIf
EndFunction

Function ApplyDeadExpressionWithoutPhoneme(Actor victim)
    if(Settings.GetEnableExpression())
        MfgExpression = Settings.GetMfgExpression()
        MfgModifierStrength = Settings.GetMfgModifierStrength()
        
        int x = 0
        while (x < 16)
            if(MfgExpression[x] > 0)
                MfgConsoleFunc.SetModifier(victim, x, MfgModifierStrength[x])
                victim.SetExpressionOverride(x, MfgExpression[x])
            endif
            x += 1
        endwhile
    endIf
EndFunction

sslBaseExpression Function RegisterDeadExpression()
    sslBaseExpression DeadExpression = SexLab.RegisterExpression("SLPGNDeadExpression")
    
    if(Settings.GetEnableExpression())
        MfgPhonemeStrength = Settings.GetMfgPhonemeStrength()
        MfgModifierStrength = Settings.GetMfgModifierStrength()
        MfgExpression = Settings.GetMfgExpression()
        
        int phases = 1
        while(phases <= 5)
            DeadExpression.AddPhase(phases,0)
            DeadExpression.AddPhase(phases,1)
            int x = 0
            while (x < 16)
                DeadExpression.SetModifier(phases, 0, x, MfgModifierStrength[x])
                DeadExpression.SetModifier(phases, 1, x, MfgModifierStrength[x])
                
                DeadExpression.SetPhoneme(phases, 0, x, MfgPhonemeStrength[x])
                DeadExpression.SetPhoneme(phases, 1, x, MfgPhonemeStrength[x])
                
                if(MfgExpression[x] > 0)
                    DeadExpression.SetMood(phases, 0, x, MfgExpression[x])
                    DeadExpression.SetMood(phases, 1, x, MfgExpression[x])
                endif
                
                x += 1
            endwhile
            phases += 1
        endwhile
    endIf
    ;DeadExpression.Save()
    Return DeadExpression
EndFunction
    

; ---------------------------------- Checking for mods and compatibility -----------------------------------------------------------------------------------------

; Checks if various mods are installed, without requiring them to be master files.
; Pretty clever, right?
; Note that I would like to put this in its own script as I feel this will grow to the largest function, and well...doesn't exactly adhere to a single responsibility principle.
; But I worry things could break if I do. I'll put that off for if or when a v3.0.0 is required. 
; I might end up compacting the plugin to make a light one, forcing a v3.0.0 but that seems a little...too much right now.
Function ModChecks()
    Form RAAS = Game.GetFormFromFile(0x800, "RASS - Visual Effects.esl") ; This checks if the plugin is even present in the first place. Since it is a light plugin, standard methods don't work here.
    Form WetAndColdMod = Game.GetFormFromFile(0x2313, "WetandCold.esp")
;   Form Facelight = Game.GetFormFromFile(0x800, "Facelight.esp")
;   Form FacelightPlus = Game.GetFormFromFile(0x800, "FacelightPlus.esp")
    Form FHU = Game.GetFormFromFile(0x3316, "sr_FillHerUp.esp")
;   Form MilkModEconomy = Game.GetFormFromFile(0x800, "milkmod.esp")
;   Form AFT = Game.GetFormFromFile(0x800, "AFT.esp")
    Form NFF = Game.GetFormFromFile(0x6950, "nwsFollowerFramework.esp") ; this is the ID for the main follower quest of NFF. 
;   Form EFF = Game.GetFormFromFile(0x800, "EFF.esp")
    Form DeathCam = Game.GetFormFromFile(0x801, "DeathCamOptions.esp") ; This is a light plugin
    Form EBTMod = Game.GetFormFromFile(0x0389EC, "dD - Enhanced Blood Main.esp") ;ID for the main "quest" of EBT.

    ; Rain, Ash and Snow Shaders
    If(RAAS)
        RAAS_BreathSpell = Game.GetFormFromFile(0x803, "RASS - Visual Effects.esl") as Spell
        main.SetRAASPresent(true)
    Else
        main.SetRAASPresent(false)
        log("RAAS not found")
    EndIf

    ; Wet and Cold
    If(WetAndColdMod)
        WetCold_Spell = Game.GetFormFromFile(0x1846, "WetandCold.esp") as Spell
        main.SetWetAndColdPresent(true)
    Else
        Main.SetWetAndColdPresent(false)
        Log("Wet and Cold not found.")
    EndIf

;   If(Facelight)
;       Set the things.
;       FacelightPresent = true
;   Else
;       FacelightPresent = false
;       Log("Facelight not found")
;   EndIf

;   If(FacelightPlus)
;       Also set the things.
;       FacelightPlusPresent = true
;   Else
;       FacelightPlusPresent = false
;       Log("Facelight Plus not found.")
;   EndIf
;
    If(FHU)
        Main.SetFHUPresent(true)
        Log("Fill Her Up found.")
        deflation.InitializeParameters()
    Else
        Log("Fill Her Up not found.")
        Main.SetFHUPresent(false)
    EndIf

;   If(MilkModEconomy)
;       Main.SetMilkModPresent(true)
;       Settings.SetMilkModCompatibility(true)
;   Else
;       Main.SetMilkModPresent(false)
;       Settings.SetMilkModCompatibility(false)
;   EndIf

    ; This assumes that they only have one framework installed..which, frankly you really should only have one. Otherwise problems may, no will arise. 

    If(NFF)
        main.SetFollowerFramework(1)
        Log("NFF Found")
    Else
        main.SetFollowerFramework(0)
        Log("NFF Not found")
    EndIf
    ;/
    If(AFT)
        main.SetFollowerFramework(2)
        Log("AFT found.")
    Else
        main.SetFollowerFramework(0)
        Log("AFT not found")
    EndIf
    /;
    ;/
    If(EFF)
        main.SetFollowerFramework(3)
        Log("EFF found.")
    Else
        main.SetfollowerFramework(0)
        Log("EFF not found.")
    EndIf
    /;

    If(DeathCam)
        main.SetDeathCamPresent(true)
        Log("Death Cam Extender found.")
    Else
        main.SetDeathCamPresent(false)
        Log("Death Cam Extender not found.")
    EndIf

    If(EBTMod)
        main.SetEBTPresent(true)
        EBTBlood_Spell = Game.GetFormFromFile(0x3300B, "dD - Enhanced Blood Main.esp") as Spell
        Log("Enhanced Blood Textures found.")
    Else
        main.SetEBTPresent(false)
        Log("Enhanced Blood Textures not found.")
    EndIf
EndFunction


; --------------------------------- Controlling fadeouts, if enabled --------------------------------------------------------------------------
Function EnableBlackScreen()
    If(Settings.GetDoFadeout())
        Log("Fade On")

        If(FadeEffect != NONE)
            FadeEffect.ApplyCrossFade()
        Else
            Game.FadeOutGame(true, true, 0.5, 3.0)
        EndIf
    EndIf
EndFunction


Function DisableBlackScreen()
    If(Settings.GetDoFadeout())
        Log("Fade Off")

        If(FadeEffect != NONE)
            ImageSpaceModifier.RemoveCrossFade()
        Else
            Game.FadeOutGame(false, true, 0.5, 1.5)
        EndIf
    EndIf
EndFunction

; --------------------------------- Repeated little code snippets. True utility functions. ---------------------------------------------------------------------------

; I put this into a function because it is used more than once. 
Function DoMenus()
    If(Settings.GetToggleMenus())
        Debug.ToggleMenus()
    EndIf
EndFunction


; First checks if Enhanced Blood Textures is installed, then it checks if the option in the MCM is enabled
Bool Function EBTCheck()
    If(main.GetEBTPresent())
        If(Settings.GetEBTNoBlood())
            return true
        Else
            return false
        EndIf
    return false
    EndIf
EndFunction


; Checks if first FHU is present, then if the option (In this case, deflating on death) is enabled.
Function DeflateCheck()
    If(Main.GetFHUPresent())
        If(Settings.GetDeflateOnDeath())
            deflation.DeathlyDeflate()
        EndIf
    EndIf
EndFunction


; Checks if first FHU is present, then if the option (In this case, deflating when full) is enabled.
Function FullDeflateCheck()
    If(Main.GetFHUPresent())
        If(Settings.GetDeflateWhenFull())
            deflation.DeflateIfFull()
        EndIf
    EndIf
EndFunction


; --------------------------------- Functions that are called on player death. ---------------------------------------------------------------------------

; These remove the breathing in the cold effects added by "Wet and Cold" , and/or "Rain, Ash And Snow shaders" on death.
Function RemoveColdEffects()
    If(Settings.GetRemoveEffects())
        If(Main.GetRAASPresent())
            If(RAAS_BreathSpell)
                PlayerRef.DispelSpell(RAAS_BreathSpell)
                PlayerRef.RemoveSpell(RAAS_BreathSpell)
            Else
                Log("Somehow, despite multiple checks, the RASS Spell was never set.")
            EndIf
        EndIf

        If(Main.GetWetAndColdPresent())
            If(WetCold_Spell)
                PlayerRef.DispelSpell(WetCold_Spell)
                PlayerRef.RemoveSpell(WetCold_Spell)
            Else
                Log("Somehow, despite checks.. The wet and cold spell was never set never set.")
            EndIf
        EndIf
    EndIf
EndFunction


Function RemoveFacelights()
    If(Main.GetFacelightType() != 0)
        PlayerRef.DispelSpell(Facelight_Spell)
        PlayerRef.DispelSpell(FacelightDark_Spell)
        PlayerRef.RemoveSpell(Facelight_Spell)
        PlayerRef.RemoveSpell(FacelightDark_Spell)
    EndIf
EndFunction


Function DisableBloodPools()
    If(EBTCheck())
        If(EBTBlood_Spell)
            PlayerRef.DispelSpell(EBTBlood_Spell)
            PlayerRef.RemoveSpell(EBTBlood_Spell)
        Else
            Log("Somehow, the EBT blood spell was never set.")
        EndIf
    EndIf
EndFunction

; --------------------------------- Functions that are called when the mod decides to give the player life again. ----------------------------------------

Function NaturalResurrection()
    ; So, what this function will do is check if a (currently present, but commented out until this function works.) option is enabled in the MCM.
    ; If it is, then we will check if the player character has any dragon souls, and if so a soul will be consumed for them to be able to live again.
    ; Maybe it'll play the soul absorbtion animation, or maybe it'll do like, the necromancer resurrection animation. I dunno.
    ; If no dragon soul is found (say, maybe a non-dragonborn character), or as an alternative it'll look for black soul gems.
    ; This will not be as flashy as a dragon soul resurrection, and will probably just be a regular resurrection / they can stay alive.
    ; Perhaps there should be an option or percentage to use either souls or gems. I dunno.

    ; If this is done via a percentage, like a  

    ; I think the resurrection / death thing is going to need to be re-worked to kill the player character, then do the resurrection options.

    ;Let's write some pseudocode for this. do I know if the code would work, no. But I'm laying the framework.

    ; If(Settings.GetNaturalResurrection())
    ; Debug.SetGodMode(true) ; Set god mode to avoid the PC dying. I believe there is something else we can use here that would make them be ignored, but that will take some playing around with.  
    ;   
        int souls = PlayerRef.GetAV("DragonSouls") as int ; Do I know if this cast is required? No. But I don't trust Papyrus or the game enough.
        If(souls > 0)
    ;       PlayAnimation(soulAbsorb)
    ;       CoolResurrection() ; I dunno. This is pseudocode after all. 
        Else
            ; Fallback to black soul gem checking.
        EndIf
    ; EndIf
EndFunction


Function AddColdEffects()
    If(Settings.GetRemoveEffects())
        If(Main.GetRAASPresent())
            PlayerRef.AddSpell(RAAS_BreathSpell, false)
        EndIf

        If(Main.GetWetAndColdPresent())
            PlayerRef.AddSpell(WetCold_Spell, false)
        EndIf
    EndIf
EndFunction


Function AddFacelights()
    If(Main.GetFacelightType() != 0)
        PlayerRef.AddSpell(Facelight_Spell, false)
        PlayerRef.AddSpell(FacelightDark_Spell, false)
    EndIf
EndFunction


Function EnableBloodPools()
    If(EBTCheck())
        PlayerRef.AddSpell(EBTBlood_Spell, false)
    EndIf
EndFunction

;------------------------------------ Debug Stuff -----------------------------

Function log(String argString)
    Debug.Trace("SLPGN Utilities: " + Utility.GetCurrentRealTime()+ " -- " + argString )    
EndFunction


; Sanity checking. Papyrus is a dumbass sometimes.
Function SetSettings()
    If(Settings == NONE)
        Settings = Game.GetFormFromFile(0x22ED, "SLPGN.esp") as SLPGN_MCMConfig
        Log("For some reason, settings did not initialise, but hopefully it has now.")
    EndIf
EndFunction

;------------------------------------ More sanity checking. -----------------------------
Function LoadLibraries()
    Form SLPGNQuest = Game.GetFormFromFile(0xD62, "SLPGN.esp")

    If(SLPGNQuest)
        Settings = Game.GetFormFromFile(0x22ED, "SLPGN.esp") as SLPGN_MCMConfig
        deflation = SLPGNQuest as SLPGNDeflateOnDeath
    EndIf
EndFunction